#!/usr/bin/perl
require("/etc/re_config.pl");
push(@INC,$modules_dir);
require("format_calls_lib.pl");
require("sublist.pl");
use Expect;
local $debug=0;
local $debug2=0;
local $exit_program=0;

sub make_gdb_form
{
	my ($operand)=@_;
	$operand=~s/^\$//;
	if($operand=~ /%/)
	{
		$operand=~ s/^%es://g;
		$operand=~ s/^%ds://g;

		if($operand=~/%.l/)
		{
			$operand=~s/%/%e/;
			$operand=~s/l$/x/;
		}
		$operand=~ s/%/\$/g;
		if($operand=~ /^\(.*\)$/)
		{
			$operand=~s/\(//g;
		}
		$operand=~s/\(/+/g;
		if($operand=~ /,/)
		{
			$operand=~s/,/+/;
			$operand=~s/,/*/;
			if($operand=~ /\)/)
			{
				#$operand=~s/\)/*4/g;
				$operand=~s/\+\+/+/g;
			}
		}
		$operand=~s/\)//g;
	}
	return $operand;
}

sub do_cmd
{
	my ($command)=@_;
	$gdb->send("$command\n");
	my ($status,$line,$prompt);
	my $status=0;
	my $last_line;
	my $line="";
	my $out_lines;
	my $out_lines_n;
	my $retstr="";
	while(1)
	{
		$patidx=$gdb->expect($timeout,
			-re => '.*\(gdb\)|^>',
			-re => 'Type <return>'
			#-re => '\r\n'
		) or die "time out\n";

		my $cur_ret_str=$gdb->exp_before;
		$retstr="$retstr$cur_ret_str";
		chomp($retstr);
		$retstr=~s/\r//g;
		if($gdb->exp_match_number==1)
		{
			if($cur_ret_str =~ /The program is not being run./ || $retstr =~ /No registers./ || $retstr =~ /Program exited / || $retstr=~/The program is not being run./)
			{
				$exit_program=1;
			}
			last;
		}elsif($gdb->exp_match_number==2)
		{
			if($debug)
			{
				print STDERR "send return continue";
			}
			my $cur_ret_str=$gdb->exp_before;
			$retstr="$retstr$cur_ret_str";
			$gdb->send("\n");
		}elsif($gdb->exp_match_number==3)
		{
			if($debug)
			{
				print STDERR "retstr-> [$command: $retstr]\n";
			}
			if($retstr =~ /The program is not being run./ || $retstr =~ /No registers./ || $retstr =~ /Program exited / || $retstr=~/The program is not being run./)
			{
				$exit_program=1;
				last;
			}
		}
	}
	if($debug2)
	{
		print STDERR "retstr-> [$command: $retstr]\n";
	}
	return $retstr;
}

sub do_cmd_n_get_last_line
{
	my ($command)=@_;
	$retstr=do_cmd($command);
	$retstr=~s/^\n//;
	@retarray=split(/\n/,$retstr);
	my $lastline=$retarray[$#retarray];
	if($debug2)
	{
		print STDERR "lastline: [$lastline]\n";
	}
	return $lastline;
}

sub is_memory_position
{
	my ($addr_line)=@_;

	if($addr_line=~/^0xbff[0-9a-f]+/ || $addr_line=~/0x80[0-9a-f][0-9a-f][0-9a-f]+/)
	{
		return 1;
	}
	return 0;
}
			
sub get_dump_line
{
	my ($level,$target_part,@cmd_lines)=@_;
	my $indent_str=get_indent_str($level);
	my $ret_line="";
	my $do_print;
	$do_print=0;
	foreach my $cmd_line (@cmd_lines)
	{
		chomp($cmd_line);
		$cmd_line=~s/\r//;
		$cmd_line=~s/\n//;
		if($do_print==1)
		{
			$ret_line="$ret_line$indent_str$cmd_line\n";
		}
		if(!$do_print && ($target_part eq "") || ($cmd_line=~ /^$target_part:/))
		{
			$do_print=1;
			$ret_line="$indent_str$cmd_line\n";
		}
	}
	return $ret_line;
}

sub dump_operand
{
	my ($memory,$level)=@_;
	my $target_part=get_addr_str($memory);
	chomp($target_part);
	my $ret_dump_line="";
	my $dump_line="";
	if(is_memory_position($target_part))
	{
		my $dump_line=do_cmd_n_get_last_line("x/a $target_part");
		my $dst_part=$dump_line;
		$dump_line=~s/^/	/;
		$dst_part=~s/^.*:[ \t]*//g;
		if(is_memory_position($dst_part))
		{
			$dump_line=dump_operand($dst_part,$level+1);
			$ret_dump_line="$ret_dump_line$dump_line";
		}else{
			my $operation=find_operation($addr);
			if($operation eq "")
			{
				$operation="x/s";
			}
	
			$dump_line=do_cmd("$operation $target_part");
			$ret_dump_line="$ret_dump_line$dump_line";
		}
	}
	return $ret_dump_line;
}

sub get_pid
{
	my ($ps_name,$except_pid)=@_;
	my $ret_pid_line="";
	if($ps_name)
	{
		while($ret_pid_line eq "")
		{
			open(PS_RESULT,"/bin/ps ax|") || die "Cannot open: ps";
			while(<PS_RESULT>)
			{
				$ps_line=$_;
				$ps_line=~s/^[ \t]*[^ ^\t]+[ \t]+[^ ^\t]+[ \t]+[^ ^\t]+[ \t]+[^ ^\t]+[ \t]+//;
				chomp($ps_line);
				$ps_line=~s/[ \t]+$//g;
				if($ps_line=~/$ps_name/ && $ps_line!~/perl/)
				{
					$ret_pid_line=$_;
					$ret_pid_line=~s/^[ \t]*//;
					$ret_pid_line=~s/[ ]+.*$//;
					chomp($ret_pid_line);
					if($ret_pid_line ne $except_pid)
					{
						last;
					}else{
						$ret_pid_line="";
					}
				}
			}
			close PS_RESULT;
			if($ret_pid_line ne "")
			{
				last;
			}
		}
	}
	return $ret_pid_line;
}



sub get_addr_str
{
	my ($str)=@_;
	$str=~s/:[ \t].*$//g;
	return $str;
}

sub get_main_str
{
	my ($str)=@_;
	$str=~s/^.*:[ \t]+//;
	return $str;
}

my $operation_table;
my $operation_table_n=0;
sub find_operation
{
	my ($addr)=@_;
	for($operation_table_i=0;$operation_table_i<$operation_table_n;$operation_table_i++)
	{
		if($operation_table[$operation_table_i][0] eq $addr)
		{
			return $operation_table[$operation_table_i][1];
		}
	}
	return "";
}

sub print_about_memory
{
	my ($addr,$operand,$level)=@_;
	$operand_str=make_gdb_form($operand);
	my $indent_str=get_indent_str($level);
	if($operand_str)
	{
		my $cmd_str=do_cmd_n_get_last_line("x/x $operand_str");
		print OUTPUTFILE "$indent_str$operand ($operand_str=$cmd_str)\n";
		$dump_str=dump_operand($cmd_str,$level+1);
		if($dump_str)
		{
			print OUTPUTFILE "$dump_str\n";
		}
	}else{
		print OUTPUTFILE "$indent_str$operand\n";
	}
}

sub get_last_line
{
	my ($line_n,@multiline)=@_;
	return $multiline[$#line_n-1];
}




use Getopt::Std;
getopt('a:b:e:o:p:s:t:');

my $fork_start='0x80571e8';
my $fork_end='0x805720b';
my $ps_name=$opt_p;
my $bin_name="";
my $break_here="0x8055e38";
my $till_here="";
my $outputfile="gdbrun.out";
my $operationfile="";
chomp($opt_a);
my $out_addr="";

if($opt_o)
{
	chomp($opt_o);
	$out_addr=$opt_o;
}

local $global_gdbconf="gdbconf";
if($opt_a)
{
	$break_here=$opt_a;
	$outputfile="$opt_a";
	if($opt_t)
	{
		$outputfile=~s/$/-$opt_t/;
	}elsif($opt_s)
	{
		$outputfile=~s/$/-$opt_s/;
	}
	$outputfile=~s/$/.gr/;
	$operationfile=$outputfile;
	$operationfile=~s/\.gdbrun$/.gdbconf/;

	$operation_table_n=0;
	foreach my $operationfile ($operationfile,$global_gdbconf)
	{
		open(OPERATIONFILE,"<$operationfile");
		while(<OPERATIONFILE>)
		{
			my ($addr,$operation)=split(/[ \t]+/);
			if($debug)
			{
				print STDERR "$addr -> $operation\n";
			}
			chomp($addr);
			chomp($operation);
			$operation_table[$operation_table_n][0]=$addr;
			$operation_table[$operation_table_n][1]=$operation;
			$operation_table_n++;
		}
		close(OPERATIONFILE);
	}
}else{
	print STDERR "usage: gdbrun.pl -a <start address> -s <stepi num> -e <except this process> -p <process_name> -b <binary_name> -t <till here> -o <end address>\n";
	print STDERR "example: gdbrun.pl -a 0x8048f94 -t 0x8049135  -b /root/tb/the-binary\n";
	print STDERR "\tthis option means from 0x8048f94 till 0x8049135 debug ./root/tb/the-binary and dump the results\n";
	exit;
}

local $no_fork_start="0x80485b0";
local $no_fork_end="0x80485b5";
$no_fork_start="";
$no_fork_end="";

chomp($opt_t);
if($opt_t)
{
	$till_here=$opt_t;
}

open(OUTPUTFILE,">$outputfile");
#OUTPUTFILE=*STDERR;
local $step_num=-1;
local $except_pid=$opt_e;
$bin_name=$opt_b;

local $gdb;
$pid_line=get_pid($ps_name,$except_pid);
print OUTPUTFILE "/usr/bin/gdb $bin_name $pid_line\n";
local $command="/usr/bin/gdb";
local @args;
local $timeout=1000;
if($pid_line)
{
	$args[0]=$bin_name;
	$args[1]=$pid_line;
}else{
	$args[0]=$bin_name;
}
local $gdb=Expect->spawn($command,@args) || die "Cannot spawn $command\n";
$gdb->log_stdout(0);

system "stty sane";
$match_patterns[0]="\(gdb\)";
$patidx=$gdb->expect($timeout,@match_patterns);
{
	my $go_on_loop=1;
	if($pid_line)
	{
		print "break *$break_here\n";
		print "cont\n";
		do_cmd_n_get_last_line("where");
		do_cmd_n_get_last_line("break *$break_here");
		do_cmd_n_get_last_line("cont");
	}else{
		if($debug)
		{
			print STDERR "break *$break_here\n";
		}
		do_cmd_n_get_last_line("break *$break_here");

		if($debug)
		{
			print STDERR "break *$fork_start\n";
		}
		do_cmd_n_get_last_line("break *$fork_start");

		$cmd_ret_line=do_cmd_n_get_last_line("command");
		if($debug)
		{
			print STDERR "set \$eip=$fork_end\n";
		}
		$cmd_ret_line=do_cmd_n_get_last_line("set \$eip=$fork_end");
		$cmd_ret_line=do_cmd_n_get_last_line("set \$eax=0");
		$cmd_ret_line=do_cmd_n_get_last_line("set \$edx=0");
		$cmd_ret_line=do_cmd_n_get_last_line("cont");
		$cmd_ret_line=do_cmd_n_get_last_line("end");

		if($no_fork_start && $no_fork_end)
		{
			if($no_fork_start)
			{
				if($debug)
				{
					print STDERR "disable 2 for $no_fork_start\n";
				}
				do_cmd_n_get_last_line("break *$no_fork_start");
				$cmd_ret_line=do_cmd_n_get_last_line("command");
				$cmd_ret_line=do_cmd_n_get_last_line("where");
				$cmd_ret_line=do_cmd_n_get_last_line("disable 2");
				$cmd_ret_line=do_cmd_n_get_last_line("cont");
				$cmd_ret_line=do_cmd_n_get_last_line("end");
			}

			if($no_fork_end)
			{
				if($debug)
				{
					print STDERR "enable 2 for $no_fork_end\n";
				}
				do_cmd_n_get_last_line("break *$no_fork_end");
				$cmd_ret_line=do_cmd_n_get_last_line("command");
				$cmd_ret_line=do_cmd_n_get_last_line("enable 2");
				$cmd_ret_line=do_cmd_n_get_last_line("cont");
				$cmd_ret_line=do_cmd_n_get_last_line("end");
			}
		}
		if($out_addr)
		{
			if($debug)
			{
				print STDERR "break *$out_addr\n";
			}
			do_cmd_n_get_last_line("break *$out_addr");
		}
		if($debug)
		{
			print STDERR "run\n";
		}
		my $cmd_ret_line=do_cmd_n_get_last_line("run");
#		while($cmd_ret_line!~/$break_here/)
#		{
#			print STDERR "[$cmd_ret_line]\n";
#			$cmd_ret_line=do_cmd_n_get_last_line("cont");
#			if($cmd_ret_line=~/The program is not being run./ || $cmd_ret_line=~/Program exited/)
#			{
#				print STDERR "continuing: [$cmd_ret_line]\n";
#				$go_on_loop=0;
#				last;
#			}
#		}
	}
	if(1 || $debug)
	{
		print STDERR "Starting...\n";
	}
	my $lastaddr=0;
	my $push_num=0;
	while($go_on_loop)
	{
		$exit_program=0;
		$step_num=-1;
		chomp($opt_s);
		if($opt_s)
		{
			$step_num=$opt_s;
		}
		$addr=0;
		if($debug)
		{
			$dump_line=do_cmd("where");
			print STDERR "where=$dump_line";
		}
		while(!$exit_program && ($step_num>0 || $step_num==-1) && $addr ne $till_here)
		{
			my $cmd_line=do_cmd_n_get_last_line("x/i \$eip");
			$addr=get_addr($cmd_line);
	
			#for information
			my $diff_addr=hex($addr)-hex($lastaddr);
			if($debug)
			{
				print STDERR "$addr $diff_addr\n";
			}
			$lastaddr=$addr;
	
			$op=get_op($cmd_line);
			$src=get_src($cmd_line);
			$dst=get_dst($cmd_line);
	
			print OUTPUTFILE "---------------------\n";
			print OUTPUTFILE "$addr: $op ";
			if($op=~ /^push/)
			{
				$push_num++;
			}elsif($op=~ /^call/)
			{
				#print_about_memory($addr,"$esp",1);
			}
			if($op=~ /^j/ || $op =~ /^call/)
			{
				print OUTPUTFILE "$src\n";
			}else{
				print OUTPUTFILE "\n";
				print_about_memory($addr,$src,1);
				if($op!~ /^mov/)
				{
					print_about_memory($addr,$dst,1);
				}
			}

			if($op=~/^call/ && $src=~/$fork_start/)
			{
				do_cmd_n_get_last_line("cont");
			}elsif($op=~/call/)
			{
				print STDERR "stepi\n";
				$retline=do_cmd_n_get_last_line("stepi 6");
				print OUTPUTFILE "$retline\n";
				print STDERR "finish\n";
				$retline=do_cmd_n_get_last_line("finish");
				print OUTPUTFILE "$retline\n";
			}else{
				do_cmd_n_get_last_line("stepi");
			}
			if($op!~/^call/ && $op!~/^push/ && $op!~/^pop/ && $op!~/^j/ && $op!~/^cmp/ && $op!~/^test/ && $op!~/^cld/)
			{
				print OUTPUTFILE "\t----->\n";
				if(0)
				{
					print_about_memory($addr,$src,1);
				}
				#if($op=~/^mov/)
				{
					print_about_memory($addr,$dst,1);
				}
			}
			if($step_num>0)
			{
				$step_num--;
			}
		}
		if(($out_addr && $ret_cmd_str =~ /$out_addr/) || $out_addr eq "")
		{
			print STDERR "last;\n";
			last;
		}
		if($go_on_loop)
		{
			$ret_cmd_str=do_cmd_n_get_last_line("cont");
		}
	}
	$gdb->send("quit\n");
}

close(OUTPUTFILE);
$gdb->hard_close();
